% Text Mining with MATLAB 2nd Edition
% www.http://www.textmininglab.net/

% CHAPTER 10: Dimensionality Reduction

disp('Example 10.1 -----------------------------------------------------');
clear;                                                              %(10.1)
load datasets_ch10a
pause;

disp('Example 10.2 -----------------------------------------------------');
% gets the indices of terms with relative frequencies above 0.1
hf_terms = (ranked_frequencies/max(ranked_frequencies))>0.1;       %(10.2a)
ranked_vocabulary(hf_terms) % Lists the terms                      %(10.2b)
% computes the percentage of running words they represent          %(10.2c)
sum(ranked_frequencies(hf_terms))/sum(ranked_frequencies)*100
pause;

disp('Example 10.3 -----------------------------------------------------');
% gets the indices of all singletons                               %(10.3a)
lf_terms = ranked_frequencies==1;    
sum(lf_terms) % compute the total amount of these terms            %(10.3b)
% computes the percentage of running words they represent          %(10.3c)
sum(ranked_frequencies(lf_terms))/sum(ranked_frequencies)*100
pause;

disp('Example 10.4 -----------------------------------------------------');
pruned_tfidfmtx = tfidfmtx(not(hf_terms|lf_terms),:);               %(10.4)
pause;

disp('Example 10.5 -----------------------------------------------------');
% converts the vocabulary string array into a character array      %(10.5a)
char_vocabulary = char(ranked_vocabulary{:});
% retains the first six characters of each term and reconverts     %(10.5b)
% the truncated character array back into a string array
stems = string(cellstr(char_vocabulary(:,1:6))');
% groups vocabulary terms into the new vocabulary of stems         %(10.5c)
[stems_vocabulary,void,pointer] = unique(stems);
length(stems_vocabulary) % gets the stem vocabulary size           %(10.5d)
pause;

disp('Example 10.6 -----------------------------------------------------');
% creates a sparse matrix of appropriate dimensions                %(10.6a)
stems_tfmtx = sparse(length(stems_vocabulary),length(verses));
% updates the term frequency for each new stem category            %(10.6b)
for k=1:length(ranked_vocabulary) 
    stems_tfmtx(pointer(k),:) = stems_tfmtx(pointer(k),:)+tfmtx(k,:);
end
pause;

disp('Example 10.7 -----------------------------------------------------');
% creates a histogram of pointers from terms to stems              %(10.7a)
nterms = hist(pointer,length(stems_vocabulary));
% identifies those stems groiping more than 8 terms                %(10.7b)
index = find(nterms>8) 
pause;

disp('Example 10.8 -----------------------------------------------------');
stems_vocabulary(index(1))                                         %(10.8a)
% gets all vocabulary terms that have been reduced to 'accept'     %(10.8b)
ranked_vocabulary(pointer==index(1))
pause;

disp('Examples 10.9 and 10.10 ------------------------------------------');
disp('THESE EXAMPLES ARE NOT MATLAB SCRIPTS')                 %(10.9-10.10)
pause;

disp('Example 10.11 ----------------------------------------------------');
hf_terms = (ranked_frequencies/max(ranked_frequencies))>0.02;     %(10.11a)
lf_terms = ranked_frequencies<=5;
pruned_tfidfmtx = tfidfmtx(not(hf_terms|lf_terms),:);             %(10.11b)
whos pruned_tfidfmtx
pause;

disp('Example 10.12 ----------------------------------------------------');
pruned_ntfidfmtx = sparse(size(pruned_tfidfmtx,1),length(verses)); %(10.12)
for k=1:length(verses) % applies L2-norm normalization
    temp = norm(pruned_tfidfmtx(:,k));
    if temp>0 % avoids divisions by zero
        pruned_ntfidfmtx(:,k) = pruned_tfidfmtx(:,k)/temp;
    end
end
pause;

disp('GENERATING FIGURE 10.1 -------------------------------------------');
cosine_scores = pruned_ntfidfmtx(:,392)'*pruned_ntfidfmtx;       
hf = figure(1);                              
set(hf,'Color',[1 1 1],'Name','Distribution of similarities with respect to verse 392');
[freq,vals] = hist(cosine_scores,20); hb = bar(vals,log(freq)+1); 
xlabel('Cosine Association Scores'); ylabel('Log-frequencies + 1');
pause;

disp('Example 10.13 ----------------------------------------------------');
[u,s,v] = svds(pruned_ntfidfmtx,100);                              %(10.13)
whos u s v
pause;

disp('GENERATING FIGURE 10.2 -------------------------------------------');
hf = figure(2);                              
set(hf,'Color',[1 1 1],'Name','100 largest singular values');
plot(diag(s),'*-'); 
xlabel('Number of Singular Value'); ylabel('Singular Value Magnitude');
pause;

disp('Example 10.14 ----------------------------------------------------');
% computes the projected document vectors in 20 dimensions  	 %(10.14.a)
projected_docs = u(:,1:20)'*pruned_ntfidfmtx;
norm_projected_docs = sparse(20,length(verses));                 %(10.14.b)
for k=1:length(verses) % applies L2-norm normalization
    temp = norm(projected_docs(:,k));
    if temp>0 % avoids divisions by zero
        norm_projected_docs(:,k) = projected_docs(:,k)/temp;
    end
end
% computes cosine similarity scores with respect to verse 392    %(10.14.c)
cosine_scores = norm_projected_docs(:,392)'*norm_projected_docs;
pause;

disp('GENERATING FIGURE 10.3 -------------------------------------------');      
hf = figure(1);                              
set(hf,'Color',[1 1 1],'Name','Distribution of similarities with respect to verse 392 in 20 dimensions');
[freq,vals] = hist(cosine_scores,20); hb = bar(vals,log(freq)+1); 
xlabel('Cosine Association Scores'); ylabel('Log-frequencies + 1');
axis([-0.2,1,0,10]);
pause;

disp('Examples 10.15 and 10.16 -----------------------------------------');
disp('THESE EXAMPLES ARE NOT MATLAB SCRIPTS')                %(10.15-10.16)
pause;

disp('Example 10.17 ----------------------------------------------------');
% applies L2-norm normalization to the rows of pruned_tfidfmtx     %(10.17)
norm_terms = sparse(size(pruned_tfidfmtx,1),size(pruned_tfidfmtx,2));
for k=1:size(pruned_tfidfmtx,1)                                       
    temp = norm(pruned_tfidfmtx(k,:)); 
    norm_terms(k,:) = pruned_tfidfmtx(k,:)/temp;
end
pause;

disp('Example 10.18 ----------------------------------------------------');
pruned_vocabulary=ranked_vocabulary(not(hf_terms|lf_terms));       %(10.18)
pause;

disp('Example 10.19 ----------------------------------------------------');
[nu,ns,nv] = svds(norm_terms,100); % performs SVD                 %(10.19a)
% computes the projections (be reminded that here we are          %(10.19b)
% projecting the rows of the matrix instead of its columns)
projected_terms = norm_terms*nv;
% applies L2-norm normalization to the projected matrix rows      %(10.19c)
for k=1:length(pruned_vocabulary)
    temp = norm(projected_terms(k,:));
    norm_projected_terms(k,:) = projected_terms(k,:)/temp;   
end
pause;

disp('Example 10.20 ----------------------------------------------------');
% gets the indices of both words                                  %(10.20a)
w1 = find(pruned_vocabulary=='star');
w2 = find(pruned_vocabulary=='shining');
% computes the number of verses in which these two words co-occur %(10.20b)
full(sum((norm_terms(w1,:)>0)&(norm_terms(w2,:)>0)))
pause;

disp('Example 10.21 ----------------------------------------------------');
full(norm_terms(w1,:)*norm_terms(w2,:)')                           %(10.21)
pause;

disp('Example 10.22 ----------------------------------------------------');
full(norm_projected_terms(w1,:)*norm_projected_terms(w2,:)')       %(10.22)
pause;

disp('Example 10.23 ----------------------------------------------------');
% enters the word pairs into two string arrays                    %(10.23a)
list1 = ["love","thirst","hungry","feet","fish"]; 
list2 = ["obedience","thirsty","poverty","knees","creature"];
% computes cosine similarity scores                               %(10.23b)
for k=1:5
    w1 = find(pruned_vocabulary==list1(k));
    w2 = find(pruned_vocabulary==list2(k));
    hd_score(k) = norm_terms(w1,:)*norm_terms(w2,:)';
    ld_score(k) = norm_projected_terms(w1,:)*norm_projected_terms(w2,:)';
end
% generates the report                                            %(10.23c)
%report = table();
%report.('Word 1') = list1';
%report.('Word 2') = list2';
%report.('HD Score') = hd_score';
%report.('LD Score') = ld_score';
%disp(report)
headers = {'Word 1','Word 2','HD Score','LD Score'};
disp(table(list1',list2',hd_score',ld_score','VariableNames',headers))
pause;

disp('Example 10.24 ----------------------------------------------------');
w = find(pruned_vocabulary=='drink'); % gets the word index       %(10.24a)
% computes and sorts similarity scores in high-dimensional space  %(10.24b)
scores = norm_terms*norm_terms(w,:)';
[sortedscores,order] = sort(scores,'descend');
hd_sc = full(sortedscores(1:10));
hd_wd = pruned_vocabulary(order(1:10))';
for k=1:10 % gets co-occurrences for 10 closest words             %(10.24c)
    temp = sum((norm_terms(w,:)>0)&(norm_terms(order(k),:)>0));
    hd_cc(k) = full(temp);
end
% computes and sorts similarity scores in low-dimensional space   %(10.24d)
scores = norm_projected_terms*norm_projected_terms(w,:)';
[sortedscores,order] = sort(scores,'descend');
ld_sc = full(sortedscores(1:10));
ld_wd = pruned_vocabulary(order(1:10))';
for k=1:10 % gets co-occurrences for 10 closest words             %(10.24e)
    temp = sum((norm_terms(w,:)>0)&(norm_terms(order(k),:)>0));
    ld_cc(k) = full(temp);
end
% generates the report                                            %(10.24f)
hds = {'HD Co-oc','HD Score','HD Words','LD Co-oc','LD Score','LD Words'};
disp(table(hd_cc',hd_sc,hd_wd,ld_cc',ld_sc,ld_wd,'VariableNames',hds))
pause;

% Generates Table 10.1 ...
wdlst = ["moon","fury","rain","cup"];
report = table();
for k=1:length(wdlst)
    w = find(pruned_vocabulary==wdlst(k));
    scores = norm_terms*norm_terms(w,:)';
    [void,order] = sort(scores,'descend');
    hd_wd = pruned_vocabulary(order(1:6))';
    report.(sprintf('HD %s',wdlst(k))) = hd_wd;
    scores = norm_projected_terms*norm_projected_terms(w,:)';
    [void,order] = sort(scores,'descend');
    ld_wd = pruned_vocabulary(order(1:6))';
    report.(sprintf('LD %s',wdlst(k))) = ld_wd;
end
disp(report) 
pause;

disp('Example 10.25 ----------------------------------------------------');
disp('THIS EXAMPLE IS NOT MATLAB SCRIPT')                          %(10.25)
pause;

disp('Example 10.26 ----------------------------------------------------');
cities = {'Atlanta','Chicago','Denver','Houston','Los Angeles','Miami',...
          'New York','San Francisco','Seattle','Washington'};     %(10.26a)
% matrix of approximate inter-city distances in kilometers     	  %(10.26b)
d = [ 0  1503  3103  1795  4956  1546  1915  5476  5586  1390
   1503     0  2355  2406  4467  3041  1825  4756  4447  1528
   3103  2355     0  2250  2127  4419  4175  2429  2614  3825
   1795  2406  2250     0  3517  2478  3635  4211  4841  3123
   4956  4467  2127  3517     0  5988  6275   888  2455  5888
   1546  3041  4419  2478  5988     0  2796  6641  6999  2363
   1915  1825  4175  3635  6275  2796     0  6582  6164   525
   5476  4756  2429  4211   888  6641  6582     0  1736  6252
   5586  4447  2614  4841  2455  6999  6164  1736     0  5962
   1390  1528  3825  3123  5888  2363   525  6252  5962     0];
pause;

disp('Example 10.27 ----------------------------------------------------');
% creates a new figure and displays a US map                      %(10.27a)
hf = figure(5); 
set(hf,'Color',[1,1,1],'Name','Placing cities into a map with MDS');
[imagen,cmap] = imread('map_usa.gif','gif');
image(imagen); colormap(cmap);
% applies metric MDS to the inter-city distance matrix in (10.26b)%(10.27b)
[coord,stress] = mdscale(d,2,'criterion','metricstress');
% rotates and rescales to fit the resulting embeding into the map %(10.27c)
rotang = pi/17; scale1 = 325; scale2 = 200; offset1 = 20; offset2 = 30;
rotcoord = -coord*[cos(rotang),-sin(rotang);sin(rotang),cos(rotang)];
minval = min(rotcoord); maxval = max(rotcoord);
normcoord(:,1) = (rotcoord(:,1)-minval(:,1))/(maxval(:,1)-minval(:,1));
normcoord(:,2) = (rotcoord(:,2)-minval(:,2))/(maxval(:,2)-minval(:,2));
mapcoord(:,1) = normcoord(:,1)*scale1 + offset1;
mapcoord(:,2) = (1-normcoord(:,2))*scale2 + offset2;
% plot city locations and names into the map                      %(10.27d)
hold on; plot(mapcoord(:,1),mapcoord(:,2),'*b'); hold off;
text(mapcoord(:,1),mapcoord(:,2)+7,cities,'Color',[0,0,1],'FontSize',10,...
     'HorizontalAlignment','center'); 
pause;

disp('Example 10.28 ----------------------------------------------------');
% Creates a string array with the selected vocabulary terms       %(10.28a)
list = ["bird","cloud","field","fish","flock","goat","lightning",...
    "mountain","rain","river","sea","sheep","sky","storm","thunder","wind"];
% Gets the corresponding index for each word in the list          %(10.28b)
for k=1:length(list)
    idx(k) = find(pruned_vocabulary==list(k));
end
% Computes the cosine distance matrix for the subset of words     %(10.28c)
thedist = pdist(full(norm_terms(idx,:)),'cosine');  
% Applies MDS to the computed cosine distance matrix              %(10.28d)
limit = statset('MaxIter',500); % Sets the maximun number of iterations
[map,stress] = mdscale(thedist,2,'criterion','stress','options',limit);
pause;

disp('Example 10.29 ----------------------------------------------------');
hf = figure(6); % Creates a new figure                            %(10.29a)
set(hf,'Color',[1,1,1],'Name','Constructing a map of words with MDS');
plot(map(:,1),-map(:,2),'*'); % Plots the word locations          %(10.29b)
axis([-0.6,0.6,-0.3,0.3]);
% Prints each word below its corresponding location               %(10.29c)
text(map(:,1),-map(:,2)-0.02,list,'HorizontalAlignment','center');
pause;

disp('Example 10.30 ----------------------------------------------------');
word_embedding = fastTextWordEmbedding                             %(10.30)
pause;

disp('Example 10.31 ----------------------------------------------------');
% Gets embedded vector representations for man, men and woman     %(10.31a)
w1 = word2vec(word_embedding,"man");
w2 = word2vec(word_embedding,"men");
w3 = word2vec(word_embedding,"woman");
% Gets the closest word in the embedding to vector men-man+woman  %(10.31b)
w4 = vec2word(word_embedding,w2-w1+w3)
pause;

disp('Example 10.32 ----------------------------------------------------');
w1 = word2vec(word_embedding,"sell");                             %(10.32a)
w2 = word2vec(word_embedding,"sold");
w3 = word2vec(word_embedding,"buy");
w4 = vec2word(word_embedding,w2-w1+w3)                            %(10.32b)
pause;

disp('Example 10.33 ----------------------------------------------------');
wds(1,:) = ["Portugal","Lisbon","Spain"];
wds(2,:) = ["dinner","p.m.","breakfast"];
wds(3,:) = ["nurses","hospital","teachers"];
for n=1:size(wds,1)
    for k=1:3, wv(k,:) = word2vec(word_embedding,wds(n,k)); end
    w4 = vec2word(word_embedding,wv(2,:)-wv(1,:)+wv(3,:));
    disp(compose("'%s' - '%s' + '%s' = '%s'",[wds(n,[2,1,3]),w4]))
end
pause;
  
disp('Example 10.34 ----------------------------------------------------');
w1 = word2vec(word_embedding,"nurses");                           %(10.34a)
w2 = word2vec(word_embedding,"hospital");
w3 = word2vec(word_embedding,"judges");
w4 = vec2word(word_embedding,w2-w1+w3,5)'                         %(10.34b)
pause;

disp('Example 10.35 ----------------------------------------------------');
% loads the dataset and creates a string array of verses          %(10.35a)
clear; load datasets_ch10a
for k=1:length(verses), docs(k) = verses(k).text; end
% preprocess the data                                             %(10.35b)
tokdocs = tokenizedDocument(docs);
tokdocs = erasePunctuation(tokdocs);
tokdocs = lower(tokdocs);
tokdocs = removeStopWords(tokdocs);
pause;

disp('Example 10.36 ----------------------------------------------------');
% trains the word embedding                                        %(10.36)
word_embedding = trainWordEmbedding(tokdocs,...
                 'Dimension',30,'MinCount',10,'NumEpochs',20)
pause;

disp('Example 10.37 ----------------------------------------------------');
words = word_embedding.Vocabulary;                                %(10.37a)
vectors = word2vec(word_embedding,words);
word_map = tsne(vectors);
hf = figure(8);                                                   %(10.37b)
set(hf,'Color',[1,1,1],'Name','Word Embedding Visualization');
textscatter(word_map,words)
pause;

% CREATES FIGURE 9 -------------------------
hf = figure(9);             
set(hf,'Color',[1,1,1],'Name','Word Embedding Neighborhood of "two"');
textscatter(word_map,words)
pause;

disp('Example 10.38 ----------------------------------------------------');
clear; load datasets_ch10b                                         %(10.38)
whos
pause;

disp('Example 10.39 ----------------------------------------------------');
% creates a string array with the raw text from books             %(10.39a)
for k=1:length(books), docs(k) = books(k).text; end
% preprocess the documents                                        %(10.39b)
tokdocs = tokenizedDocument(docs);
tokdocs = erasePunctuation(tokdocs);
tokdocs = lower(tokdocs);
% creates a bag-of-words model and computes the tfidf matrix      %(10.39c)
bow = bagOfWords(tokdocs);
bow = removeInfrequentWords(bow,5);
tfidfmtx = tfidf(bow);
size(tfidfmtx)
pause;

disp('Example 10.40 ----------------------------------------------------');
% creates a document embedding by means of SVD                    %(10.40a)
[u,s,v] = svd(full(tfidfmtx'));
document_embbeding = u(:,1:200)'*tfidfmtx';
% creates a 2D map of the embedding with MDS                      %(10.40b) 
docdist = pdist(full(document_embbeding'),'cosine');  
limit = statset('MaxIter',500); 
[map,stress] = mdscale(docdist,2,'criterion','sammon','options',limit);
pause;

disp('Example 10.41 ----------------------------------------------------');
% gets information about titles and categories                    %(10.41a)
for k=1:length(books), 
    titles(k) = books(k).name; 
    groups(k) = books(k).group;
end
hf = figure(10); % creates the plot                               %(10.41b)
set(hf,'Color',[1,1,1],'Name','Document Embedding Visualization');
textscatter(map,titles,'TextDensityPercentage',100,'ColorData',groups)
axis([-1,1,-1,1])
pause;


